/*
|	FILE: easydtd.c
|
|			    Copyright 1993, 1994, 1995 
|                               All Rights Reserved
|				Norman E. Smith, CDP
|
|	This program reads a file of tag names and generates a
|	skeleton DTD. Only elements are included at this time. Other
|	DTD constructs will be considered in future editions.
|
|	The input file has one tag name per line in an outline
|	type form. Each level of indention is used as a clue to
|	determine the content model.
|
|	Rules of the game include:
|	1)  The first time an element is used, the nesting must
|	    be specified. All later occurrances will be dropped.
|	2)  <ENTITIES>, <ELEMENTS>, and <ATTRIBUTES> commands MUST start
|	    in column 1!
|	3)  #PCDATA must be in all caps when used
|	4)  Very little syntax checking is done!
|	6)  Attribute list must fit all on one line.
|
|	This is an initial, proof of concept program. There are many
|	improvements in mind for future versions.
|
|	Future improvements include:
|	1) Documentation
|	2) check that the first char must be alpha
|	3) Figure out how to handle multiple grouping
|	4) Add "include" files
|
|	DATE:
| V2.1.1
| 	10/14/94	Folded in changes for EasyDTD to compile under
|			Unix. From Warren Baird.
| V2.2
|	10/20/94	Made multi-line Entities be OK. Now must make sure
|			they wrap on output.[nes]
|	10/21/94	Added code to wrap entity lines. [nes]
|			Added code to write skeleton AWK programs to substitute
|			entities automatically. [nes]
|	10/22/94	Added code to write skeleton OmniMark programs. [nes]
|	10/23/94	Added -c flag to include default SGML declaration. [nes]
| V2.3
|	02/09/95	Started changes for the new attribute processing. [nes]
|	02/10/95	Got first working version of attribute processing.
|	02/11/95	Added validation of attribute value and defaults.
|	02/16/95	Modified AWK code to put out both types of
|			programs. Started Data Dictionary code.
|	02/22/95	Finished first pass at the Data Dict. Code.[nes]
|	02/28/95	Fixed bug in Data Dict. code handline of attributes. [nes]
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef __TURBOC__
#include <dos.h>
#endif

#define FOREVER		for(;;)
#define TRUE		1
#define FALSE		0
#define LSIZE		2
#define ASCII_BLANK	' '
#define TAB		'\t'

#define EMPTY		0
#define PCDATA		1
#define GROUP		2
#define MULTIPLE	3
#define ELEMENT		4
#define ATTRIBUTE	5
#define ENTITY		6
#define GENERAL		7
#define PARAMETER	8
#define COMMENT		9
#define INCLUDE		10
#define APP_FILE	11
#define DTD		12
#define ENTITY_ONLY	13
#define OMNIMARK	14
#define DATA_DICT	15
#define SIMPLE_ATTR	16
#define COMPLEX_ATTR	17
#define NEW_ATTR	18
#define SGML_DEC        19
#define AWK		20
#define AWK2		21

#define ERROR		255
#define MAX_CONTENT	4096
#define MAX_CONN	3
#define MAX_SEP		3
#define BUFSIZE		4096
#define RIGHT_MARGIN	70
/***********************+-----------------+
|			| DATA STRUCTURES |
|			+-----------------+
*/
	/*
	| This structure contains data necessary to build
	| <!ELEMENT statements, including content models.
	*/
struct element{
	int  type;
	int  level;
	int  len;
	char *name;
	char *cm;			/* Stuff to build content model */
	char *separator;
	char *occurance;
	char *inclusion;		/* Inclusions, exclusions, or comments*/
	char *content_model;		/* not used yet */
	struct comment_lines *comments;
		/*
		| The att_list is for future use. Each element should
		| have the list of attributes attached directly to
		| it. This initial implementation takes the simpler
		| approach and just builds a list of attributes.
		*/
	struct attribute *att_list;
	struct element   *next_element;
}    ;
	/*
	| This entity holds an entity and its definitiosn.
	*/
struct entity{
	long type;
	int  len;
	char *name;
	char *value;
	struct comment_lines *comment;
	struct entity *next_entity;
}    ;
	/*
	| This data structure holds attributes. Each attribute is
	| a separate entry. The attributes should really be attached
	| directly to its element.
	*/
struct attribute{
	long len;
	char *element;		/* Element name this attribute is attached to */
	char *name;		/* Attribute name                             */
	char *values;		/* Allowable values                           */
	char *default_value;	/* Default value                              */
	struct comment_lines *comment;
	struct attribute     *next_attribute;
}    ;
	/*
	| The data_line structure holds each line in the input file
	| in memory. That way each line is read into a linked list of
	| lines where it is then processed by the program as if
	| reading from a file.
	*/
struct data_line{
	long len;
	char *line;
	struct data_line *prev_line;
	struct data_line *next_line;
}    ;
struct comment_lines{
	char *comment;
	struct comment_lines *next_comment;
} ;
/***********************+------------------+
|			| GLOBAL VARIABLES |
|			+------------------+
*/
#ifdef _BCC_
unsigned _stklen = 16284;		/* Increased stack size in BC++ */
#endif

int  verbose         = FALSE;
int  quiet           = FALSE;
int  list_elem       = FALSE;
int  long_names      = FALSE;
int  short_ref       = FALSE;
int  entity_file     = FALSE;
int  entity_elements = FALSE;
int  minimize        = FALSE;
int  tag_width       = 8;
int  attr_val_width  = 8;
int  record_num      = 0;
int  wrap            = 20;
int  type_output     = DTD;
int  SGML_declaration= FALSE;
int  attr_state      = SIMPLE_ATTR;

struct element *element0;
struct element *elementn;

struct entity *entity0;
struct entity *entityn;

struct data_line *line0;
struct data_line *linen;
struct data_line *next_data_line;

struct attribute *attribute0;
struct attribute *attributen;

FILE *fd_in;
FILE *fd_out;
FILE *fd_entity;

char *got_eof               = FALSE;

char ptitle[]               = "EASYDTD V2.3";
char pdate[]                = __DATE__;
char prev_tag[256]	    = "";
char minimization[32]       = " - - ";
char default_separator[32]  = ",";
char default_occurrance[32] = "";
char default_attribute[32]  = "#IMPLIED";
char default_attr_value[32] = "CDATA";
char connectors[MAX_CONN+1] = ",&|";
char separators[MAX_SEP+1]  = "*?+";

char file_in[256];
char file_out[256];
char file_entity[64] = "entity.ent";

char in_buff[BUFSIZE];
char out_buff[BUFSIZE];

/********************************************************************
=====================================================================
=                      VALIDATION TABLES                            =
=                                                                   =
=====================================================================
*********************************************************************/
char *attr_val_table[]={
	"CDATA",    "ENTITY",  "ENTITIES",
	"ID",       "IDREF",   "IDREFS",
	"NAME",     "NAMES",   "NMTOKEN",
	"NMTOKENS", "NUTOKEN", "NUTOKENS",
	"cdata",    "entity",  "entities",
	"id",       "idref",   "idrefs",
	"name",     "names",   "nmtoken",
	"nmtokens", "nutoken", "nutokens",
	""
	};
char *attr_default_table[]={
	"#CONREF",  "#CURRENT", "#FIXED",
	"#IMPLIED", "#REQUIRED",
	"#conref",  "#current", "#fixed",
	"#implied", "#required",
	""
	};
char *content_model_table[]={
	"#PCDATA", "CDATA", "RCDATA",
	"ANY",     "EMPTY",
	""
	};
/********************************************************************
=====================================================================
=                      FNSPLIT FUNCTIONS                            =
=                                                                   =
= This section of code is a generic implementation of fnsplit()     =
= found in BC++.  This code is taken from Until with a few minor    =
= changes.                                                          =
=                                                                   =
=====================================================================
*********************************************************************/

#ifndef BCC

#define EXT     0
#define FNAME	1
#define DIR	2
#define DDRIVE	3
#define SOS     4
#define DONE	5

/***********************+-------------+
|			| reverse_str |
|			+-------------+
*/
int reverse_str(char *string)
{
    int  i;
    int  len;
    int  truth = FALSE;
    char *orig = string;
    char temp[260];      /* 255 plus slop */

    len       = strlen(string);
    if(len > 255){
        truth = TRUE;
        return(truth);
    }
    string += len - 1;
    for(i=0;i<len;i++){
        temp[i] = *string--;
    }
    temp[i] = '\0';
    strcpy(orig,temp);
    return(truth);
}
/***********************+------------+
|			| proc_fname |
|			+------------+
*/
int proc_fname(char *f_name,char ch)
{
	int  F_state = FNAME;

	if(ch == '\\' || ch == '/' || ch == ']' || ch == '['){
		*f_name   = '\0';	/* close off string          */
		F_state   = DONE;
		return(F_state);
	}
	if(ch == ':'){
		*f_name   = '\0';	/* close off string          */
		F_state   = DDRIVE;
		return(F_state);
	}
	*f_name = ch;
	return(F_state);
}
/***********************+----------+
|			| proc_ext |
|			+----------+
*/
int proc_ext(char *ext,char ch)
{
	int  F_state = EXT;

	if(ch == '.'){
		F_state = DONE;
		*ext++  = ch;
		*ext    = '\0';
		return(F_state);
	}
	if(ch == '\\' || ch == '/' || ch == ']' || ch == '['){
		*ext     = '\0';	/* close off string          */
		F_state  = DIR;
		return(F_state);
	}
	if(ch == ':'){
		*ext     = '\0';	/* close off string          */
		F_state  = DDRIVE;
		return(F_state);
	}
	*ext = ch;
	return(F_state);
}
/***********************+---------+
|			| fnsplit |
|			+---------+
*/
void fnsplit(char *fname,char *drive,char *dir,char *file,char *ext)
{
	int  i;
	int  len;
	int  F_state;
	char ch;

	char *ptr_drive;
	char *ptr_dir;
	char *ptr_file;
	char *ptr_ext;

	ptr_drive = drive;
	ptr_dir   = dir;
	ptr_file  = file;
	ptr_ext   = ext;
/*
strcpy(fname,"c:\\dir\\file.ext");
printf("fnsplit:\n");
fnsplit(fname,drive,dir,file,ext);
printf("%s\n",ext);
printf("%s\n",file);
printf("%s\n",dir);
printf("%s\n",drive);
*/
	drive[0]  = '\0';
	dir[0]    = '\0';
	file[0]   = '\0';
	ext[0]    = '\0';
	len       = strlen(fname);
	i         = 0;
	F_state   = EXT;
        len--;				/* backup to last char in string */
		/*
		| Verify that something was passed....
		*/
	if(len < 0){
		printf("fn_split: Empty input string \n");
		return;
	}
/*******************************************************************
|			Handle File Extension
|*******************************************************************
*/
		/*
		| Loop looking for file extension
		*/
	FOREVER{
		ch = *(fname + len);		/* Get next char  */

		switch(F_state){
		case EXT:
			F_state          = proc_ext(ptr_ext,ch);
			ptr_ext++;
			if(F_state      == DONE){
				F_state  = FNAME;
			}
			if(F_state      == DIR){
				strcpy(file,ext);	/* no extension-> file name  */
			}
			len--;
			if(len < 0){		/* Test for SOS              */
				F_state  = SOS;
				*ptr_ext = '\0';	/* close off string          */
				strcpy(file,ext);	/* no extension-> file name  */
				ptr_ext  = ext;
				ext[0]   = '\0';
			}
			break;
/*******************************************************************
|			Handle File Name
|*******************************************************************
*/
		case FNAME:
			F_state = proc_fname(ptr_file,ch);
			ptr_file++;
			if(F_state      == DONE){
				F_state  = DIR;
				len++;
			}
			len--;
			if(len < 0){		/* Test for SOS              */
				F_state  = SOS;
				*ptr_file = '\0';	/* close off string          */
				ptr_file  = ext;
/*
				file[0]   = '\0';
*/
			}
			break;
/*******************************************************************
|			Handle Directory Name
|*******************************************************************
*/
		case DIR:
			if(ch == ':'){
				dir[i]   = '\0';	/* close off string          */
				i        = 0;
				F_state  = DDRIVE;
			}else{
				dir[i++] = ch;
				len--;
				if(len < 0){		/* Test for SOS              */
					F_state  = SOS;
					dir[i]   = '\0';	/* close off string          */
				}
			}
			break;
/*******************************************************************
|			Handle Disk Drive Name
|*******************************************************************
*/
		case DDRIVE:
			drive[i++] = ch;
			len--;
			if(len < 0){		/* Test for SOS              */
				F_state    = SOS;
				drive[i]   = '\0';	/* close off string          */
			}
			break;
/*******************************************************************
|			Handle Start of String
|*******************************************************************
*/
		case SOS:
			printf("fn_split: Invalid F_state value!!\n");
		}
		if(F_state == SOS){
			break;
		}
	}
		/*
		| Strings must be mirrored before done...
		*/
	reverse_str(ext);
	reverse_str(file);
	reverse_str(dir);
	reverse_str(drive);
}

#endif

/********************************************************************
=====================================================================
=                      UTILITY FUNCTIONS                            =
=                                                                   =
= This section of code contains the code to read the input file     =
= memory for processing. It also contains other misc. functions that=
= are use in other parts of the program.                            =
=====================================================================
*********************************************************************/
/***********************+-----------+
|			| my_malloc |
|			+-----------+
| Used only for extreme cases of debugging....
*/
char* my_malloc(int size)
{
	char *value;

	value = (char*) malloc(size);
	printf("V: %lu\n",(unsigned long)value);
	return(value);
}
/***********************+--------------+
|			| if_separator |
|			+--------------+
| This function returns TRUE when the character is a separator character.
| Otherwise, it returns FALSE.
*/
long if_separator(char ch){
	long rvalue = FALSE;
	int  i;

	for(i=0;i<MAX_SEP;i++){
		if(separators[i] == ch){
			rvalue    = TRUE;
			break;
		}
	}
	return(rvalue);
}
/***********************+--------------+
|			| if_connector |
|			+--------------+
| This function returns TRUE when the character is a separator character.
| Otherwise, it returns FALSE.
*/
int if_connector(char ch){
	long rvalue = FALSE;
	int  i;

	for(i=0;i<MAX_CONN;i++){
		if(connectors[i] == ch){
			rvalue    = TRUE;
			break;
		}
	}
	return(rvalue);
}
/***********************+----------------+
|			| read_data_line |
|			+----------------+
| This function reads the next line in from the input file. The
| max line length is 256 bytes.
*/
char* read_data_line()
{
	int  i;
	int  len;
	char *status;
	char *last_char;
	char ch;

	status = fgets(in_buff,BUFSIZE,fd_in);
	if(!status){			/* EOF */
		return(status);
	}
		/*
		| get rid of trailing \n
		*/
	len           = strlen(in_buff);
	last_char     = &in_buff[len-1];
	if(*last_char == '\n'){
		in_buff[len-1] = '\0';
	}
		/*
		| Skip blank lines
		*/
ReadAgain:
	if(!strlen(in_buff)){
		status = read_data_line();
		if(!status){
			return(status);	     /* End of file...             */
		}
		goto ReadAgain;		     /* Read until a non-empty line*/
	}		/*
			| Process comment line. I chose goto instead
			| of recursion here to prevent problems that
			| could be caused by a bunch of comment lines
			| in a row.
			*/
	if(in_buff[0] == ';'){
		in_buff[0] = '\0';
		goto ReadAgain;
	}
		/*
		| Test for tabs. They cause major problems!
		| Bail out rather than try to recover...
		*/
	for(i=0;i<len-1;i++){
		ch = in_buff[i];
		if(ch == TAB){
			printf("Tab character detected in line: \n");
			printf("%s\n",in_buff);
			exit(0);
		}
	}
	return(status);
}

/***********************+-------------+
|			| create_line |
|			+-------------+
| The first element is special because it does not link to
| a previous entry.
| This function creates an data_line data structure, fills in what
| it can, but does NOT link it into the chain of lines.
*/
struct data_line* create_line(char *string)
{
	int               len;
	char              *tname;
	struct data_line  *temp;
		/*
		| Allocate memory to hold the line
		*/
	temp             = (struct data_line*)malloc(sizeof(struct data_line));
	if(!temp){
		printf("Could not malloc() - temp\n");
		printf("  %s\n",string);
		exit(0);
	}
	tname            = (char*)malloc(strlen(string)+2);
	if(!tname){
		printf("Could not malloc() - tname\n");
		printf("  %s\n",string);
		exit(0);
	}

	strcpy(tname,string);
	temp->line         = tname;
	temp->len          = strlen(tname);
	temp->next_line    = 0;
	temp->prev_line    = 0;
	return(temp);
}
/***********************+------------+
|			| load_line0 |
|			+------------+
*/
void load_line0()
{
		/*
		| Create the initial line in the list
		*/
	linen            = 0;
	line0            = create_line(" ");
	linen            = line0;
	line0->line      = 0;
	line0->prev_line = 0;
}
/***********************+-----------+
|			| load_line |
|			+-----------+
*/
void load_line()
{
	struct data_line *new_data_line;
	struct data_line *prev_line;
		/*
		| The steps to adding a new line in the linked list are:
		| 1)
		| 2) Create & populate the entity data structure
		| 3) Link the new entity into the list
		*/
	prev_line                = linen;
	new_data_line            = create_line(in_buff);
	new_data_line->prev_line = prev_line;
	new_data_line->next_line = 0;
	prev_line->next_line     = new_data_line;
	linen                    = new_data_line;
}
/***********************+---------------------+
|			| read_next_data_line |
|			+---------------------+
| This function logically reads the next line in the input stream. It
| really returns the next line from the data_line linked list. It
| returns the address of the string or 0 for EOF and copies the
| line into in_buff[].
*/
char* read_next_data_line()
{
	char *status = got_eof;

	if(got_eof){
		return((char*) 0);
	}
	if(!linen->next_line){
		got_eof = (char*)TRUE;
	}
	strcpy(in_buff,linen->line);
	status = linen->line;
	linen  = linen->next_line;
	return(status);
}
/********************************************************************
=====================================================================
=                          ENTITIES                                 =
=                                                                   =
= This section of code contains the Entity processing.              =
=====================================================================
*********************************************************************/
/***********************+---------------+
|			| lookup_entity |
|			+---------------+
*/
struct entity *lookup_entity(char *name)
{
	long          truth = FALSE;
	struct entity *this_ent;

	int           status;

	this_ent = entity0;		/* start at the beginning */
	do{
			/*
			| Only make full compare when the length
			| matches.
			*/
		if(this_ent->len     == strlen(name)){
			status        = strcmp(this_ent->name,name);
			if(!status){
				truth = TRUE;
				break;
			}
		}
		if(!this_ent->next_entity){
			break;
		};
		this_ent = this_ent->next_entity;
	}while(this_ent->next_entity);
		/*
		| The above loop does not test the last value
		*/
	if(this_ent->len == strlen(name)){
		status     = strcmp(this_ent->name,name);
		if(!status){	truth = TRUE;	}
	}
	if(!truth){
		this_ent = 0;
	}
	return(this_ent);
}
/***********************+---------------+
|			| create_entity |
|			+---------------+
| The first element is special because it does not link to
| a previous entry.
| This function creates an element data structure, fills in what
| it can, but does NOT link it into the chain of elements.
*/
struct entity* create_entity(char *ename, char *evalue, long etype)
{
	int            len;
	char           *tname;
	char           *tvalue;
	struct entity  *temp;
	struct comment_lines *tcomment;
		/*
		| Check that tag name is <= tag_width
		*/
	if(strlen(ename) > tag_width){
		fprintf(stderr,"create_entity: Element name too long: %s\n",ename);
	}
	temp             = (struct entity*)malloc(sizeof(struct entity));
	if(!temp){
		printf("create_entity cannot alloc memory for entity\n");
		exit(0);
		}
	tname            = (char*)malloc(strlen(ename)+2);
	if(!tname){
		printf("create_entity cannot alloc memory for ename\n");
		exit(0);
		}
	tvalue           = (char*)malloc(strlen(evalue)+2);
	if(!tvalue){
		printf("create_entity cannot alloc memory for tvalue\n");
		exit(0);
	}
/*
	tcomment         = (char*)malloc(sizeof(struct comment_lines));
	if(!tcomment){
		printf("create_entity cannot alloc memory for tcomment\n");
		exit(0);
	}
*/
	strcpy(tname,ename);
	strcpy(tvalue,evalue);
	temp->name         = tname;
	temp->len          = strlen(tname);
	temp->type         = etype;
	temp->value        = tvalue;
	temp->comment      = 0;
	temp->next_entity  = 0;
	return(temp);
}
/***********************+-------------+
|			| link_entity |
|			+-------------+
*/
void link_entity(struct entity *new_ent)
{
	entityn->next_entity = new_ent;
	entityn              = new_ent;
}
/***********************+--------------+
|			| parse_entity |
|			+--------------+
| The input is from in_buff. The type is returned as
| either GENERAL or PARAMETER. The entity name is in ent_name and
| the entity value in value. Note: ent_name and value MUST have
| been allocated before calling parse_entity!!
|
| This implementation assumes the entity fits all on one line!
*/
long parse_entity(char *ent_name, char *value)
{
	int  i;
	int  len;
        int  collected_EOS;
	long param          = GENERAL;
	char *s_in_buff     = in_buff;
	char *status;
	char delimiter;
	char ch;
	char string[1024] = "";
		/*
		| Skip leading blanks in the entity name
		*/
	len = strlen(s_in_buff);
	for(i=0;i<len;i++){
		ch     = *s_in_buff;
		if(ch != ASCII_BLANK){
			break;
		}
		s_in_buff++;
	}
		/*
		| Test for parameter entity
		*/
	if(*s_in_buff == '%'){
		param  = PARAMETER;
		s_in_buff++;
	}
		/*
		| now parse up to the first whitespace and put
		| it into the entity name.
		*/
	len = strlen(s_in_buff);
	for(i=0;i<len;i++){
		ch     = *s_in_buff++;
		if(ch == ASCII_BLANK){
			break;
		}
		string[i] = ch;
	}
	string[i] = '\0';
	strcpy(ent_name,string);
	string[0] = '\0';
		/*
		| The value starts at the first " or ' and ends
		| at the corresponding " or ', even if it crosses
		| line boundaries!
		*/
		/*
		| Skipping to first " or '
		*/
	i              = 0;
	while(*s_in_buff){
		ch     = *s_in_buff++;
		if((ch == '"') || (ch == '\'')){
			string[i++]   = ch;
			delimiter     = ch;	/* Save the delimiter char */
			break;
		}
	}
	string[i] = '\0';
		/*
		| Now extract the entity value...
		*/
	collected_EOS  = FALSE;
Collect_String:
	while(*s_in_buff){
		ch     = *s_in_buff++;
		if(ch == delimiter){
			string[i++]   = ch;
			collected_EOS = TRUE;
			break;
		}
		string[i++] = ch;
	}
	if(!collected_EOS){			/* Not found end of string, so read another line*/
		status = read_next_data_line();
		if(status){			/* Not EOF */
			s_in_buff     = in_buff;
			goto Collect_String;
		}
	}
	string[i] = '\0';

	strcpy(value,string);
	return(param);
}
/***********************+--------------+
|			| load_entity0 |
|			+--------------+
*/
void load_entity0()
{
	long type;
	char ent_name[64];
	char value[1024];
		/*
		| Create the initial entity in the list
		*/
	type    = parse_entity(ent_name,value);
	entity0 = create_entity(ent_name,value,type);
	entityn = entity0;
}
/***********************+-------------+
|			| load_entity |
|			+-------------+
*/
void load_entity()
{
	long type;
	char ent_name[64];
	char value[1024];
	struct entity *new_ent;
	struct entity *status;
		/*
		| Test for first entity; if the first
		| then do special processing and get out...
		*/
	if(entity0 == 0){
		load_entity0();
		return;
	}
		/*
		| PUT CODE TO HANDLE COMMENTS/INCLUDES HERE FOR ENTITIES
		*/
		
		/*
		| The steps to adding a new entity in the linked list are:
		| 1) Parse the input fields
		| 2) Create & populate the entity data structure
		| 3) Link the new entity into the list
		*/
/*	ename    = calc_level(&level);	*/
	type    = parse_entity(ent_name,value);
		/*
		| Look up the name in the list and see what to
		| do
		*/
	status   = lookup_entity(ent_name);
	if(!status){		/*
				|  If name does not exist, go ahead
				|  and create a new list element
				*/
		new_ent = create_entity(ent_name,value,type);
		link_entity(new_ent);
	}else{			/*
				| Entity already exists! Do something...
				| ( This is handled administratively
				|   in this version. Later the program
				|   may attempt to figure out which
				|   entry to keep.)
				| The definition is kept for grouping
				| purposes, but not written.
				*/
		if(!quiet){
			fprintf(stderr,"Duplicate entity detected! ");
			fprintf(stderr,"Using first definition of %s\n",ent_name);
		}
	}
}
/********************************************************************
=====================================================================
=                          ATTRIBUTES                               =
=                                                                   =
= This section of code contains Attribute processin functions.      =
=====================================================================
*********************************************************************/
/***********************+------------------+
|			| create_attribute |
|			+------------------+
| This function creates an attribute data structure, fills in what
| it can, but does NOT link it into the chain of elements.
*/
struct attribute* create_attribute(char *ename, char *avalue,
	char *value, char *value_default)
{
	int              len;
	char             *tname;
	char             *tvalue;
	char             *value_tmp;
	char             *default_tmp;
	struct attribute *temp;

	temp             = (struct attribute*)malloc(sizeof(struct attribute));
	if(!temp){
		printf("create_attribute cannot alloc memory for temp\n");
		exit(0);
	}
	tname            = (char*)malloc(strlen(ename)+2);
	if(!tname){ printf("create_attribute cannot alloc memory for tname\n");
		exit(0);
	}
	tvalue           = (char*)malloc(strlen(avalue)+2);
	if(!tvalue){ printf("create_attribute cannot alloc memory for tvalue\n");
		exit(0);
	}

	value_tmp        = (char*)malloc(strlen(value)+2);
	if(!value_tmp){
		printf("create_attribute cannot alloc memory for value_tmp\n");
		exit(0);
	}

	default_tmp      = (char*)malloc(strlen(value_default)+2);
	if(!default_tmp){
		printf("create_attribute cannot alloc memory for default_tmp\n");
		exit(0);
	}

	strcpy(tname,ename);
	strcpy(tvalue,avalue);
	strcpy(value_tmp,value);
	strcpy(default_tmp,value_default);
	temp->element        = tname;
	temp->len            = strlen(tvalue);
	temp->name           = tvalue;
	temp->values         = value_tmp;
	temp->default_value  = default_tmp;
	temp->comment        = 0;
	temp->next_attribute = 0;
	return(temp);
}
/***********************+----------------+
|			| link_attribute |
|			+----------------+
*/
void link_attribute(struct attribute *new_attr)
{
	attributen->next_attribute = new_attr;
	attributen                 = new_attr;
}
/***********************+---------------+
|			| extract_token |
|			+---------------+
*/
char* extract_token(char *string, char *value, char token)
{
	char ch;

	FOREVER{
		ch        = *string++;
		if(ch    == token){  break;  }
		if(!ch){	/* return 0 for EOL  */
			*string   = 0;
			*value    = '\0';
			return(string);
		}
		*value++   = ch;
	}
	if(token != ' '){		/* handle ) */
		*value++ = ch;
	}
	*value = '\0';
	return(string);
}
/***********************+---------------+
|			| extract_paren |
|			+---------------+
*/
char* extract_paren(char *string, char *value)
{
	string = extract_token(string,value,(char)')');
	return(string);
}
/***********************+-------------------+
|			| extract_attribute |
|			+-------------------+
*/
char* extract_attribute(char *string, char *value)
{
		/*
		| Element name in the attribute statment is the
		| first token in the line is the element name to
		| attach the attributes to.
		*/
	string = extract_token(string,value,(char)' ');
	return(string);
}
/***********************+---------------------+
|			| skip_leading_blanks |
|			+---------------------+
| Remove leading whitespace from a string. The address of the
| first non-whitespace char is returned on success or 0 on
| failure.
*/
char* skip_leading_blanks(char *string)
{
	char ch;

	FOREVER{
		ch        = *string;
		if(ch   != ' '){  break;  }
		if(!ch){
			string = (char *)0;
			break;  
		}
		string++;
	}
	return(string);
}
/***********************+-----------------------+
|			| validate_lookup_table |
|			+-----------------------+
| Truth = match, False = no match
| Note: errors must be reported by the calling program.
*/
int validate_lookup_table(char *table[], char *value)
{
	int  i;
	int  truth;
	int  status;

	i     = 0;
	truth = FALSE;
	do{
		status = strcmp(value,table[i]);
		if(!status){		/* got a match */
			truth = TRUE;
			break;
		}
		i++;
	}while(table[i]);
	return(truth);
}
/***********************+--------------------+
|			| extract_attr_value |
|			+--------------------+
| Extract the value portion of the attribute into 'value'.
*/
char *extract_attr_value(char *string, char *value)
{
	int  truth;
		/*
		| skip leading blanks
		*/
	if(*string   == ' '){
		string = skip_leading_blanks(string);
		return(string);
	}
		/*
		| Value is either single word constant or a
		| string in ().
		*/
	if(*string == '('){
		string = extract_paren(string,value);
		return(string);
	}
		/*
		| Not a set of values, so get next token and
		| validate it.
		*/
	string = extract_attribute(string,value);
		/*
		| Lookup up value in table
		*/
/*
| Should an invalid value be changed to a valid default value or
| should the default field be checked and the two reversed later???
| For right now, just report it and keep going.
*/
	truth = validate_lookup_table(attr_val_table,value);
	if(!truth){				/* false means not found   */
		printf("extract_attr_value:\n");
		printf("  Table: attr_val_table\n");
		printf("  Value: %s\n",value);
	}
	return(string);
}
/***********************+----------------------+
|			| extract_attr_default |
|			+----------------------+
*/
char *extract_attr_default(char *string, char *default_value)
{
	int truth;
		/*
		| skip leading blanks
		*/
	if(*string   == ' '){
		string = skip_leading_blanks(string);
	}
		/*
		| Consider rest of line to be the default value
		*/
	strcpy(default_value,string);
		/*
		| Validate it.
		*/
	truth = validate_lookup_table(attr_default_table,default_value);
	if(!truth){				/* false means not found   */
		if(*default_value != '"'){	/* " Strings not an error  */
			printf("extract_attr_default:\n");
			printf("  Table: attr_default_table\n");
			printf("  Value: %s\n",default_value);
		}
	}
	return(string);
}
/***********************+--------------+
|			| extract_name |
|			+--------------+
| Extract the element name from an attribute line. Leading whitespace
| means this line is a continuation. The element name is put in name
| if this is a new attribute record. 
|
| The address of the first non-whitespace char is returned for simple
| attributes or 0 for complex attributes.
*/
char* extract_name(char *att_line, char* name)
{
	char ch;
		/*
		| NOTE: If the line contains a leading blank, then the line
		|	is a continuation of the previous attribute.
		*/
	if(*att_line   == ' '){
		att_line = skip_leading_blanks(att_line);
		return(att_line);
	}
		/*
		| Element name in the attribute statment is the
		| first token in the line is the element name to
		| attach the attributes to.
		*/
	attr_state        = SIMPLE_ATTR;	/* Assume the simple form     */
	FOREVER{
		ch        = *att_line++;
		if(ch    == ASCII_BLANK){  break;  }
		if(!ch){  
			attr_state = NEW_ATTR;
			att_line--;		/* back up 1 char so can test again */
			break;
		}
		*name++   = ch;
	}
	*name = '\0';
	att_line = skip_leading_blanks(att_line);
	if(!att_line){				/* Nothing past name means complex.*/
		attr_state = NEW_ATTR;
	}else{
		if(!strlen(att_line)){		/* double check for name only*/
			attr_state = NEW_ATTR;
		}
	}
	return(att_line);
}
/***********************+----------------+
|			| lookup_element |
|			+----------------+
| This is really an element related function. It is placed in the
| middle of the attribute code to prevent compile errors.
*/
struct element *lookup_element(char *name)
{
	long            truth = FALSE;
	struct element *this_elem;

	int             status;

	this_elem = element0;		/* start at the beginning */
	do{
			/*
			| Only make full compare when the length
			| matches.
			*/
		if(this_elem->len == strlen(name)){
			status     = strcmp(this_elem->name,name);
			if(!status){
				truth = TRUE;
				break;
			}
		}
		if(!this_elem->next_element){
			break;
		};
		this_elem = this_elem->next_element;
	}while(this_elem->next_element);
		/*
		| The above loop does not test the last value
		*/
	if(this_elem->len == strlen(name)){
		status     = strcmp(this_elem->name,name);
		if(!status){	truth = TRUE;	}
	}
	if(!truth){
		this_elem = 0;
	}
	return(this_elem);
}
/***********************+---------------------+
|			| process_simple_attr |
|			+---------------------+
| Extract and process all of the attributes on line for a simple
| attribute.
*/
void process_simple_attr(char *att_line, char *ename)
{
	struct attribute *new_attr;
	char aname[128];

	FOREVER{
		att_line = extract_attribute(att_line,aname);
		if(!(*aname)){	break;  }		/* end of line */
		if(attribute0 == 0){
			attribute0 = create_attribute(ename,aname,
				default_attr_value, default_attribute);
			attributen = attribute0; 
		}else{
			new_attr = create_attribute(ename,aname,
				default_attr_value, default_attribute);
			link_attribute(new_attr);
		}
	}
}
/***********************+----------------------+
|			| process_complex_attr |
|			+----------------------+
*/
void process_complex_attr(char *att_line, char *ename)
{
	struct attribute *new_attr;
	char aname[256];
	char values[256];
	char val_def[256];

	aname[0]   = '\0';
	values[0]  = '\0';
	val_def[0] = '\0';
		/*
		| Extract attribute name
		*/
	att_line = extract_attribute(att_line,aname);
	if(!(*aname)){
		printf("process_complex_attr: ");
		printf("No attribute name specified!\n");
		return;
	}
		/*
		| Extract attribute value, if any
		*/
	att_line = extract_attr_value(att_line,values);
	if(!(*values)){		/* No values specified, fill in defaults*/
		if(!strcmp(aname,"ID") &&
			!strcmp(aname,"id")){	/* Element Name is ID   */
			strcpy(values,"ID");
		}else{
			if(!strcmp(aname,"IDREF") &&
				!strcmp(aname,"idref")){ /* Element Name is IDREF   */
				strcpy(values,"IDREF");
			}else{
				strcpy(values,default_attr_value);
			}
		}
	}
		/*
		| Extract default value, if any
		*/
	att_line = extract_attr_default(att_line,val_def);
	if(!(*val_def)){	/* No default specified, take default   */
		strcpy(val_def,default_attribute);
	}

	if(attribute0 == 0){
		attribute0 = create_attribute(ename,aname,
			default_attr_value, default_attribute);
		attributen = attribute0;
	}else{
		new_attr = create_attribute(ename,aname,values,val_def);
		link_attribute(new_attr);
	}
}
/***********************+-----------------+
|			| load_attribute0 |
|			+-----------------+
*/
char* load_attribute0(char *att_line, char *ename)
{
	char *status;
	struct element *estatus;
	char aname[256];
		/*
		| Create the initial element in the list
		*/
	att_line   = extract_attribute(att_line,aname);
	if(!aname){  return((char*)0);  }
	attribute0 = create_attribute(ename,aname,(char*)0, (char*)0);
	attributen = attribute0;
	return(att_line);
}
/***********************+----------------+
|			| load_attribute |
|			+----------------+
*/
void load_attribute()
{
	char *att_line = in_buff;
	struct element   *estatus;
	struct attribute *status;

	static char ename[256];
		/*
		| PUT CODE TO HANDLE COMMENT/INCLUDE HERE
		*/
		/*
		| The steps to adding a new attribute in the linked list are:
		| 1) Parse off the element name, parse attribute name...
		| 2) Create & populate the element data structure
		| 3) Link the new element into the list
		*/
	att_line  = extract_name(att_line,ename);
	if(attr_state == NEW_ATTR){	/* Entity name, but no attr to process*/
		attr_state = COMPLEX_ATTR;
		return;
	}
		/*
		| Look up the name in the list and see what to
		| do about the level #.
		*/
	estatus   = lookup_element(ename);
	if(!estatus){	/*
			|  If name does not exist,
			|  report it, and keep going.
			*/
		printf("Element %s used with an attribute, ",ename);
		printf("but has not been defined yet\n");
	}
		/*
		| Test for first element; if the first
		| then do special processing and get out...
		*/
	switch(attr_state){
	case SIMPLE_ATTR:
		process_simple_attr(att_line,ename);
		break;
	case COMPLEX_ATTR:
		process_complex_attr(att_line,ename);
		break;
	default:
		printf("load_attribute: Invalid attr_state!\n");
	}
}
/********************************************************************
=====================================================================
=                          ELEMENTS                                 =
=                                                                   =
= This section of code contains the functions necessar to process   =
= the <ELEMENTS> command.                                           =
=====================================================================
*********************************************************************/
/***********************+-------------+
|			| validate_cm |
|			+-------------+
*/
int validate_cm(char *cm_string)
{
	int  valid     = TRUE;
/*	int  occurs    = FALSE;  */
	int  connector = FALSE;
	char ch;

	FOREVER{
		ch = *cm_string++;
		if(!ch) { break; }
		switch(ch){
		case '*':		/* Occurance indicators     */
		case '?':
		case '+':
/*			occurs = TRUE;
*/
					/* 
					| A connector cannot come before
					| a separator!
					*/
			if(connector){
				printf("validate_cm: An occurs character cannot appear after a connector!\n");
				valid = FALSE;
				return(valid);
			}
			break;
		case '(':		/* Grouping chars.          */
		case ')':
			connector = FALSE;  /* Grouping chars reset connector rule */
			break;
		case ',':		/* Separators               */
		case '&':
		case '|':
			connector = TRUE;
			break;
		default:
			printf("/nvalidate_cm: invalid char in model group!\n");
			valid = FALSE;
			return(valid);
/*			break;   */
		}
	}
	return(valid);
}
/***********************+---------------+
|			| extract_ename |
|			+---------------+
| This module takes an element line in ename, extracts it
| into ename. This means that ename should be a copy of the full
| line!! Otherwise, the content model and inclusions will be
| lost!
*/
void extract_ename(char *ename,char *cm_string)
{
	int            i;
	int            j            = 0;
	int            len;
	int            exit;
	int            status;
	char           *tname       = ename;
	char           string[256];
	char           c_string[32] = "";
	char           ch;
		/*
		| loop through ename looking for end of tag name
		*/
	exit = FALSE;
	len  = strlen(tname);
	for(i=0;i<len;){
		if(exit){  break; }
		ch = *tname++;
		switch(ch){
		case ',':
		case '|':
		case '&':
		case '?':
		case '+':
		case '*':
		case ' ':
/*		case '(':	*/
		case ')':
			exit = TRUE;
			break;
		default:
			if(ch == '('){
				c_string[j++] = ch;;
			}else{
				string[i++]   = ch;		/* Element name ends up in string[] */
			}
		}
	}
	string[i]   = '\0';
	c_string[j] = '\0';
	strcpy(ename,string);
	strcpy(cm_string,c_string);
}
/***********************+-------------+
|			| parse_ename |
|			+-------------+
| This module takes an element line in ename, separates it into
| the element name, the content model info, and inclusions/exclusions.
*/
void parse_ename(char *ename, char *cm, char* inclusion)
{
	int            i;
	int            j = 0;
	int            len;
	int            exit;
	int            status;
	char           *tname = ename;
	char           string[256];
	char           cm_string[64] = "";	/* Content Model work str */
	char           i_string[64]  = "";	/* Inclusion work string  */
	char           ch;
/*
| Parse off separators and occurrance from name
| and load into correct fields within the
| element structure.
*/
		/*
		| Get the element name
		*/
	strcpy(string,ename);
	extract_ename(string,cm_string);
		/*
		| Put rest of ename into content model work area
		| adjust to skip elem. name
		*/
	j     = strlen(cm_string);
	tname = tname + strlen(string) + j;
	len   = strlen(tname);
	for(i=0;i<len;i++){
		ch = *tname++;
		switch(ch){
		case ',':
		case '|':
		case '&':
		case '?':
		case '+':
		case '*':
		case '(':
		case ')':
			cm_string[j++] = ch;
			break;
		default:
			goto DoneCM;
/*			break;  */
		}
	}
DoneCM:
	cm_string[j] = '\0';		/* Content model chars end up in cm_string[] */
	status = validate_cm(cm_string);
	if(!status){			/* An error... */
		printf("  %s\n",ename);
	}
		/*
		| At this point, we are at the end of the content model
		| info. If there are no inclusions, control will fall
		| through the loop without affecting anything else.
		| Otherwise, the inclusion is collected in i_string[].
		*/
	for(j=0;i<len;i++){
		ch            = *tname++;
		i_string[j++] = ch;
	}
	i_string[j] = '\0';
		/*
		| NOTE: It is up to the calling program to check the
		|       length of the tag name!
		*/
	strcpy(ename,string);
	strcpy(cm,cm_string);
	strcpy(inclusion,i_string);
}
/***********************+----------------+
|			| create_element |
|			+----------------+
| The first element is special because it does not link to
| a previous entry.
| This function creates an element data structure, fills in what
| it can, but does NOT link it into the chain of elements.
|
| ename		Element name.
| elevel	Element level #.
*/
struct element* create_element(char *ename, int elevel)
{
	int            i,j;
	int            len;
	char           *tname;
	char           *itemp;
	struct element *temp;
/*	char           *xxx = ename;	*/
	char           sep[8];
	char           occur[8];
	char           string[256];
	char           cm_string[64] = "";
	char           inclusion[64] = "";
	char           ch;

	strcpy(string,ename);
	parse_ename(string,cm_string,inclusion);

	if(list_elem){
		printf("%s\n",string);
	}

	if(strlen(cm_string) > 15){
		printf("create_element: cm_string length > 15\n");
		exit(0);
	}
		/*
		| Allocate memory for all the pieces of the element
		| data structure.
		*/
	temp             = (struct element*)malloc(sizeof(struct element));
	if(!temp){
		printf("create_element cannot alloc memory for element\n");
		printf("element: %s\n",string);
		printf("Rec#   : %d\n",record_num);
		exit(0);
	}
	itemp            = (char*)malloc(strlen(inclusion) + 2);
	if(!itemp){
		printf("create_element cannot alloc memory for inclusion\n");
		printf("inclusion: %s\n",inclusion);
		printf("Rec#   : %d\n",record_num);
		exit(0);
	}
	tname            = (char*)malloc(strlen(string)+4);
	if(!tname){
		printf("create_element cannot alloc memory for tname\n");
		printf("tname: %s\n",string);
		printf("Rec# : %d\n",record_num);
		exit(0);
	}
	temp->cm         = (char*)malloc(32);
	if(!temp->cm){
		printf("create_element cannot alloc memory for cm\n");
		printf("tname: %s\n",string);
		printf("Rec# : %d\n",record_num);
		exit(0);
	}
	temp->separator  = (char*)malloc(16);
	if(!temp->separator){
		printf("create_element cannot alloc memory for separator\n");
		printf("tname: %s\n",string);
		printf("Rec# : %d\n",record_num);
		exit(0);
	}
	temp->occurance  = (char*)malloc(16);
	if(!temp->occurance){
		printf("create_element cannot alloc memory for occurance\n");
		printf("tname: %s\n",string);
		printf("Rec# : %d\n",record_num);
		exit(0);
	}
		/*
		| Now, fill the structure...
		*/
	strcpy(tname,string);
	temp->name         = tname;
	temp->len          = strlen(tname);
	temp->type         = EMPTY;
	temp->level        = elevel;
	temp->next_element = 0;
	temp->att_list     = 0;
	temp->inclusion    = itemp;
	strcpy(temp->cm,cm_string);
	strcpy(temp->inclusion,inclusion);
	strcpy(temp->occurance,default_occurrance);
	strcpy(temp->separator,default_separator);
	return(temp);
}
/***********************+------------+
|			| calc_level |
|			+------------+
| This function looks at the current input buffer and calculates
| the level number for this element. 
*/
char* calc_level(int *level)
{
	int  i;
	int  l;
	struct element *new_ele;
	char ch;

	l      = 0;
	i      = 0;
	*level = 0;
		/*
		| Search for first non blank, keeping track of the
		| indention. i should point to the first non-blank
		| char in the line.
		*/
	FOREVER{
		ch     = in_buff[i++];
		if(ch == ASCII_BLANK){
			l++;
		}else{
			i--;
			break;
		}
	}
	*level  = l / LSIZE;		/* calc real level based on # chars/indent */
	return(&in_buff[i]);
}
/***********************+--------------+
|			| link_element |
|			+--------------+
*/
void link_element(struct element *new_elem)
{
	elementn->next_element = new_elem;
	elementn               = new_elem;
}
/***********************+---------------+
|			| load_element0 |
|			+---------------+
*/
void load_element0()
{
	int  level = 0;
	char *ename;

		/*
		| Create the initial element in the list
		*/
	ename    = calc_level(&level);
	element0 = create_element(ename,level);
	elementn = element0;
}
/***********************+--------------+
|			| load_element |
|			+--------------+
*/
void load_element()
{
	int  truth;
	int  level = 0;
	char *ename;
	struct element *new_elem;
	struct element *status;
	char string[64];
/*	char i_string[64]  = "";    */
	char cm_string[32] = "";
		/*
		| Test for first element; if the first
		| then do special processing and get out...
		*/
	if(element0 == 0){
		load_element0();
		return;
	}
		/*
		| PUT CODE TO HANDLE COMMENTS/INCLUDE HERE
		*/
		/*
		| The steps to adding a new element in the linked list are:
		| 1) figure out the indention level
		| 2) Create & populate the element data structure
		| 3) Link the new element into the list
		*/
	ename    = calc_level(&level);
		/*
		| Look up the name in the list and see what to
		| do about the level #. First, parse off the element
		| name and look that up.
		*/
	strcpy(string,ename);
	extract_ename(string,cm_string);
		/*
		| Check that tag name is <= tag_width
		*/
	if((long_names) && (strlen(string) > tag_width)){
		fprintf(stderr,"parse_ename: Element name too long: %s\n",string);
	}
	status   = lookup_element(string);
	if(!status){		/*
				|  If name does not exist, go ahead
				|  and create a new list element
				*/
		new_elem = create_element(ename,level);
			/*
			| #PCDATA is not written out as an element,
			| but must be kept in the table for the
			| content model to be handled properly.
			*/
		truth = validate_lookup_table(content_model_table,ename);
		if(truth){
			new_elem->type = MULTIPLE;
		}
		link_element(new_elem);
	}else{			/*
				| Element already exists! Do something...
				| ( This is handled administratively
				|   in this version. Later the program
				|   may attempt to figure out which
				|   entry to keep.)
				| The definition is kept for grouping
				| purposes, but not written.
				*/
		new_elem       = create_element(ename,level);
		new_elem->type = MULTIPLE;
		link_element(new_elem);

		if(!quiet){
			fprintf(stderr,"Duplicate element detected! ");
			fprintf(stderr,"Using first definition of %s\n",string);
		}
	}
}
/********************************************************************
=====================================================================
=                          GENERAL CODE                             =
=                                                                   =
= This is the main body of the program and functions that are       =
= necessary for any of the data types to be processed.              =
=====================================================================
*********************************************************************/
/***********************+-----------------------+
|			| lookup_elem_attribute |
|			+-----------------------+
| Note: A bad side effect of the way this is coded is that when
|	the end of list is signaled, you are on the last record,
|	not past the last record. So, the calling program must use
|	this_att instead of the return value.
*/
struct attribute* lookup_elem_attribute(char *element, struct attribute *this_att)
{
	FOREVER{
		if(!this_att->next_attribute){	/* Got end of attributes */
			return((struct attribute*)0);
		}
		if(!strcmp(this_att->element,element)){	/* found it... */
			return(this_att);
		}
		this_att = this_att->next_attribute;
	}
}
/***********************+----------------+
|			| read_next_line |
|			+----------------+
| This function logically reads the next input line. The physical file
| has already been read into memory at startup. read_next_line
| simply reads the next line from the linked list. The return
| matches the old read_next_line (now read_data_line) so the rest of
| program is not affected...
*/
char* XX_read_next_line()
{
	int  len;
	char *status;
	char *last_char;

	status = read_data_line();
	return(status);
}
char* read_next_line()
{
	int  len;
	char *status;
	char *last_char;

	record_num++;
	status = read_next_data_line();
	return(status);
}
/***********************+------------+
|			| dump_linen |
|			+------------+
*/
void dump_linen()
{
	int i = 0;
	char *not_done;

	FOREVER{
	   not_done = read_next_line();
	   if(not_done){
		printf("%d %s\n",i++,in_buff);
	   }else{
		break;
	   }
	}
}
/***********************+--------------+
|			| init_easydtd |
|			+--------------+
*/
void init_easydtd()
{
	char *ename;
	char *not_done;

	printf("\n%s\n",ptitle);
	printf("%s\n\n",pdate);

	if ((fd_in = fopen(file_in, "r")) == NULL){
		fprintf(stderr, "Cannot open input file.\n");
		fprintf(stderr, "  %s\n",file_in);
		exit(1);
	}

	if ((fd_out = fopen(file_out, "w")) == NULL){
		fprintf(stderr, "Cannot open output file.\n");
		fprintf(stderr, "  %s\n",file_out);
		exit(1);
	}

	if(entity_file){
		if ((fd_entity = fopen(file_entity, "w")) == NULL){
			fprintf(stderr, "Cannot open entity file.\n");
			exit(1);
		}
	}

	element0       = 0;
	elementn       = 0;
	entity0        = 0;
	entityn        = 0;
	attribute0     = 0;
	attributen     = 0;
	line0          = 0;
	linen          = 0;
	next_data_line = linen;
		/*
		| Load entire file into memory...
		*/
	load_line0();
	FOREVER{
	   not_done = read_data_line();
	   if(not_done){
		load_line();
	   }else{
		break;
	   }
	}
	fclose(fd_in);
	linen      = line0->next_line;
	got_eof    = FALSE;
	record_num = 0;
}
/***********************+--------------+
|			| exit_easydtd |
|			+--------------+
*/
void exit_easydtd()
{
	if(short_ref){
		fputs("\n           <!-- Entities for Short References -->\n",fd_out);
		fputs("<!ENTITY   ptag   STARTTAG \"p\"         -- Paragraph start-tag -->\n",fd_out);
		fputs("<!SHORTREF docmap                      -- Map for general use --\n",fd_out);
		fputs("		  \"&#RS;&#RE;\" ptag    -- Blank line is <p> -->\n",fd_out);
		fputs("<!USEMAP   docmap \%doctype;>\n",fd_out);
	}
	fputs("]>\n",fd_out);
}
/***********************+-------+
|			| usage |
|			+-------+
*/
void usage()
{
	printf("\n%s\t%s\n",ptitle,pdate);
	printf("\n");
	printf("Usage:   easydtd [-flags] file_in file_out\n\n");
	printf("Where:\n");
	printf("flags are:\n");
	printf("-A	 Output .APP file instead of DTD \n");
	printf("-C	 Default SGML Decl.\n");
	printf("-E	 Show elements.\n");
	printf("-G	 Gen. Entity content models\n");
	printf("-K	 Gen. OmniMark Program\n");
	printf("-L	 Flag long names\n");
	printf("-M	 End tag minimization\n");
	printf("-N	 Gen. Entity Elements\n");
	printf("-O'cc'	 Set default occurs to cc \n");
	printf("-P	 Include <p> short ref\n");
	printf("-Q	 Quiet mode.\n");
	printf("-S'cc'	 Set default separator to cc \n");
	printf("-T'nn'	 Set max tag len to n\n");
	printf("-V	 verbose mode.\n");
	printf("-W	 Gen. AWK Entity Replacement Prog.\n");
	printf("-Y       Gen. Data Dictionary\n");
	printf("-X       Write Entity entries only\n");
	printf("-Z	 Gen. AWK Replacement Prog.\n");
	printf("Press <Enter> to continue...");
#ifdef _BCC_
        getch();
#else
        getchar();
#endif
        printf("\n");
	printf("file_in  is the input filename\n");
	printf("file_out is the output filename\n\n");
	printf("\n\n\n\n\n\n\n\n\n\n");
	printf("\n");
	printf("EasyDTD is Copyrighted 1994, 1995\n");
	printf("All Rights Reserved\n");
	printf("License granted for unlimited use of this executable program.\n\n");
	printf("See EASYDTD.LIC for license and EASYDTD.DOC for documentation.\n\n");
	printf("Norman E. Smith\n");
	printf("72745.1566@Compuserve.com\n\n");
}
/***********************+---------------+
|			| dump_elements |
|			+---------------+
*/
void dump_elements()
{
	struct element *this_elem;

	if(element0 == 0){   return;  }
	printf("\nELEMENTS:\n");
	printf("Level\tName\n");
	this_elem = element0;
	FOREVER{
		printf("%d\t%s \n",this_elem->level,this_elem->name);
		if(!this_elem->next_element){  break;  }
		this_elem = this_elem->next_element;
	}
}
/***********************+---------------+
|			| dump_entities |
|			+---------------+
*/
void dump_entities()
{
	struct entity *this_ent;

	if(entity0 == 0){  return; }
	printf("\nENTITIES:\n");
	this_ent = entity0;
	FOREVER{
		printf("%s %s\n",this_ent->name,this_ent->value);
		if(!this_ent->next_entity){  break;  }
		this_ent = this_ent->next_entity;
	}
}
/***********************+-----------------+
|			| dump_attributes |
|			+-----------------+
*/
void dump_attributes()
{
	struct attribute *this_att;

	if(attribute0 == 0){  return; }
	printf("\nATTRIBUTES:\n");
	printf("Elem\tAttrib\n");
	this_att = attribute0;
	FOREVER{
		printf("%s\t%s\n",this_att->element,this_att->name);
		if(!this_att->next_attribute){  break;  }
		this_att = this_att->next_attribute;
	}
}
/********************************************************************
=====================================================================
=                             DTD CODE                              =
=                                                                   =
= The following functions relate to writing the skeleton DTD.       =
=====================================================================
*********************************************************************/
/***********************+------------------+
|			| make_elem_entity |
|			+------------------+
*/
void make_elem_entity(char *name, char *content)
{
	long status;
	long len;
	char string[512];
		/*
		| Skip %doctype;
		*/
	status = strcmp(name,"%doctype;");
	if(!status){
		return;
	}
	len  = strlen(content);
	if(*(content + len - 1) == ','){
		*(content + len - 1) = '\0';
	}
	len  = strlen(name);
	if(*(name + len - 1) == ';'){
		*(name + len - 1) = '\0';
	}
	strcat(content,")");
	strcpy(string,"<!ENTITY % ");
	strcat(string,++name);
	strcat(string," \"");
	strcat(string,content);
	strcat(string,"\" >\n");
	fputs(string,fd_entity);
}
/***********************+--------------+
|			| wrap_content |
|			+--------------+
*/
void wrap_content(char* content)
{
	int  len;
	int  i;
	int  w_flag;
	char ch;
	char buff[128] = "";
				/*
				| Check to see if string is too long and
				| write it and start over if it is...
				*/
	len  = strlen(content);
	if(*(content + len - 1) == ','){
		*(content + len - 1) = '\0';
	}
	strcat(content,")");
	len  = strlen(content);
	if(len > wrap){	/* wrap and start over */
		wrap   = 60;
		i      = 0;
		w_flag = FALSE;
		FOREVER{
			ch = *content++;
			if(!ch){   break;  }
			switch(ch){		/* Can break line on any of */
			case ',':		/*   these chars.           */
			case '|':
			case '&':
				if(i>wrap){
					w_flag = TRUE;
				}
				buff[i++]  = ch;
				break;
			default:
				buff[i++]  = ch;
			}
			if(w_flag){		/* Time to wrap it... */
				w_flag = FALSE;
				buff[i++] = '\n';
				buff[i++] = '\0';
				fputs(buff,fd_out);
				strcpy(buff,"                              ");
				i = 30;
			}
		}
		if(i){				/* output last chunk   */
			buff[i++] = '\0';
			fputs(buff,fd_out);
		}

	}else{			/* No need to wrap...  */
		fputs(content,fd_out);
	}
	return;
}
/***********************+------------------+
|			| append_next_elem |
|			+------------------+
| This function appends the next element to the content model. It
| processes the explicit content model info or takes the default
| if none is specified.
*/
void append_next_elem(struct element *next_elem, char *content)
{
	int  i;
	int  len;
	int  status;
	char *temp;
	char out_cm[16];	/*
				| out_cm collects ( that must be
				| placed before the element name
				| in the content model.
				*/
	char ch;
		/*
		| Handle elements that have no content model info
		*/
	if(!strlen(next_elem->cm)){
		strcat(content,next_elem->name);
		len = strlen(next_elem->name);
		if((*next_elem->name == '%') &&
			(*(next_elem->name + len - 1) != ';'))
		{
			strcat(content,";");
		}
			/*
			| #PCDATA does not have occurrance
			| operator attached!
			*/
		status = validate_lookup_table(attr_val_table,next_elem->name);
		if(!status){		/* Not a reserved name  */
			strcat(content,next_elem->occurance);
		}

		strcat(content,next_elem->separator);
		return;
	}else{
			/*
			| At this point, a content model entry must be
			| built based on the info stored with the element
			*/
		temp = next_elem->cm;
		i    = 0;
		FOREVER{
			ch                  = *temp++;
			if(ch == '('){
				out_cm[i++] = ch;
			}else{
				out_cm[i]   = '\0';
				temp--;
				break;
			}
		}
		strcat(content,out_cm);
		strcat(content,next_elem->name);
		len = strlen(next_elem->name);
		if((*next_elem->name == '%') &&
			(*(next_elem->name + len - 1) != ';'))
		{
			strcat(content,";");
		}
		strcat(content,temp);
	}
}
/***********************+----------------------+
|			| create_content_model |
|			+----------------------+
| This function takes the element passed to it and constructs a
| content model in the 'content' string. 'content' must have been
| allocated in the calling program. The resulting content model is
| written to disk.
*/
long create_content_model(struct element *this_elem,char *content)
{
	long           rvalue = FALSE;
	struct element *next_elem;
	struct element *dup_elem;
	int            this_level;
	int            orig_level;
	static int     i;
	char           *status;
	char           ch;

	dup_elem   = this_elem;
	next_elem  = this_elem->next_element;
	this_level = this_elem->level;
/*	orig_level = this_level;  */
		/*
		| If the next tag is same level as this one,
		| then the element is #PCDATA
		*/
	if(this_elem->next_element){
		if( next_elem->level <= this_level){
				/*
				| This tag is type #PCDATA if it
				| only occurs once
				*/
			if(this_elem->type != MULTIPLE){
				strcpy(content,"(#PCDATA)");
				return(rvalue);
			}
		}
	}else{		/*
			| Handle last element in the list
			*/
		strcpy(content,"(#PCDATA)");
		return(rvalue);
	}
		/*
		| loop, adding tags to content model until
		| the level matches again.
		*/
	this_level++;
	strcpy(content,"(");
	FOREVER{
			/*
			| If same level, add this elment to the
			| content model
			*/
		if(verbose){
			printf("%d\t%d\t%s \n",
				i++,next_elem->level,next_elem->name);
		}
		if(next_elem->level == this_level){
			append_next_elem(next_elem,content);
				/*
				| Make sure content model still ends
				| in an occurs char
				*/
			ch = *(content + strlen(content) -1);
			if(!if_connector(ch)){
				strcat(content,default_separator);
			}
			next_elem    = next_elem->next_element;
		}else{  /* Stop when level gets back in sync */
				/*
				| Time to get out of loop
				*/
			if(next_elem->level < this_level){
				break;
			}
			next_elem    = next_elem->next_element;
		}
		if(!next_elem || !next_elem->next_element){  break; }
	}
		/*
		| Handle last element in model
		*/
	if(next_elem && next_elem->level == (this_elem->level+1)){
		append_next_elem(next_elem,content);
	}
		/*
		| Check to see if getting close to a full buffer
		*/
	if(strlen(content) > MAX_CONTENT - 64){
		printf("create_content_model: Content Model too big!\n");
		printf("element: %s\n",this_elem->name);
		exit(0);
	}else{
		if((entity_file) && (*dup_elem->name == '%')){
			make_elem_entity(dup_elem->name,content);
		}
	}
	return(rvalue);
}
/***********************+-------------------+
|			| build_element_rec |
|			+-------------------+
| >>> need to check TYPE to handle COMMENTS, AND MAYBE INCLUDES
*/
void build_element_rec(struct element *this_elem)
{
	int  i;
	int  col = tag_width + 14;
	int  len;
	long error;
	char content[MAX_CONTENT];

	if(!(strcmp(this_elem->name,"#PCDATA"))){
		return;		/* Don't write #PCDATA element. */
	}
	if(!(strcmp(this_elem->name,"EMPTY"))){
		return;		/* Don't write EMPTY element. */
	}
	if(!(strcmp(this_elem->name,"%doctype;"))){
		goto Process;		/* Don't skip this one... */
	}
	if(*this_elem->name == '%'){
		if(!entity_elements){
			error = create_content_model(this_elem,content);
			return;		/* Don't write entity elements. */
		}
		len = strlen(this_elem->name);
		if(*(this_elem->name + len - 1) != ';'){
			strcat(this_elem->name,";");
		}
	}
Process:
	strcpy(out_buff,"<!ELEMENT ");
	strcat(out_buff,this_elem->name);
		/*
		| Tab minimization over so everything lines up
		*/
	len  = strlen(this_elem->name);
	len  = col - (len + 8);
	for(i=0;i<len;i++){
		strcat(out_buff," ");
	}
		/*
		| Check to see if the content model is 'EMPTY'
		| and force proper minimization if it is.
		*/
	if(!strcmp(this_elem->next_element->name,"EMPTY")){
		strcat(out_buff," - o ");
	}else{
		strcat(out_buff,minimization);
	}

	fputs(out_buff,fd_out);
	out_buff[0] = '\0';
	error       = create_content_model(this_elem,content);
		/*
		| Check content model for the reserved words
		*/
	if((strcmp(content,"(#PCDATA)"))){		/* content != (#PCDATA) */
		if(!strcmp(content,"(EMPTY,")){		/* Content = EMPTY    */
			strcpy(content,"EMPTY");
		}else{
		if(!strcmp(content,"(CDATA,")){		/* = CDATA     */
			strcpy(content,"CDATA");
		}else{
		if(!strcmp(content,"(RCDATA,")){	/* = RCDATA     */
			strcpy(content,"RCDATA");
		}else{
		if(!strcmp(content,"(ANY,")){		/* = ANY     */
			strcpy(content,"ANY");
		}else{					/* All others end up here */
			wrap_content(content);
		}}}}
	}
	if((verbose) && (error)){
		fprintf(stderr,"Content model wrapped...\n");
		fprintf(stderr,"  Element: %s\n",this_elem->name);
	}
		/*
		| Handle special cases of element names
		*/
	if(!strcmp(content,"(#PCDATA)")){
		strcat(out_buff,content);
	}
	if(!strcmp(content,"CDATA")){
		strcat(out_buff,content);
	}
	if(!strcmp(content,"EMPTY")){
		strcat(out_buff,content);
	}
	if(!strcmp(content,"RCDATA")){
		strcat(out_buff,content);
	}
	if(!strcmp(content,"ANY")){
		strcat(out_buff,"ANY");
	}
	strcat(out_buff," ");
	strcat(out_buff,this_elem->inclusion);
	strcat(out_buff," >\n");
	fputs(out_buff,fd_out);
}
/***********************+------------------+
|			| write_dtd_header |
|			+------------------+
*/
void write_dtd_header()
{
	struct element *this_elem;
		/*
		| Write out DTD file header. The first element is
		| The document type as well as the first element.
		*/
	if(this_elem != NULL){
		this_elem = element0;
		fputs("<!DOCTYPE ",fd_out);
		fputs(this_elem->name,fd_out);
		fputs("  [ \n",fd_out);
		fputs("<!-- \n",fd_out);
		fputs("     Skeleton DTD created by EASYDTD. \n\n",fd_out);
		fputs("     CAUTION: This DTD probably needs editing before use!!!\n",fd_out);
		fputs("-->\n",fd_out);
	}
}
/***********************+-----------------+
|			| wrap_entity_rec |
|			+-----------------+
*/
void wrap_entity_rec()
{
	int  i;
	int  j;
	int  len;

	char ch;
	char string[128];

	len = strlen(out_buff);
	for(i=0,j=0;i<len;i++){
		ch          = out_buff[i];
		string[j++] = ch;
		if((j > RIGHT_MARGIN) &&
				((ch == ' ') ||
				 (ch == ',') ||
				 (ch == '|')
				)){
			string[j++] = '\n';
			string[j]   = '\0';
			fputs(string,fd_out);
			j           = 24;
#ifdef _BCC_
			strnset(string,' ',j);
#endif
		}
	}
		/*
		| Write last line
		*/
	string[j]   = '\0';
	fputs(string,fd_out);
}
/***********************+------------------+
|			| build_entity_rec |
|			+------------------+
*/
void build_entity_rec(struct entity *this_ent)
{
	int  i;
	int  col = tag_width + 14;
	int  len;

	strcpy(out_buff,"<!ENTITY ");
	if(this_ent->type == PARAMETER){
		strcat(out_buff,"\% ");
	}
	strcat(out_buff,this_ent->name);
		/*
		| Tab over so everything lines up
		*/
	len  = strlen(this_ent->name);
	len  = col - (len + 8);
	if(this_ent->type == PARAMETER){ len -= 2; }
	for(i=0;i<len;i++){
		strcat(out_buff," ");
	}
	strcat(out_buff,this_ent->value);
	strcat(out_buff," >\n");
	if(strlen(out_buff) > RIGHT_MARGIN){
		wrap_entity_rec();
	}else{
		fputs(out_buff,fd_out);
	}
}
/***********************+----------------+
|			| write_entities |
|			+----------------+
*/
void write_entities()
{
	struct entity  *this_ent;
	struct element *this_elem;

	this_ent  = entity0;
	this_elem = element0;

	if(this_elem != NULL){
		fputs("\n                <!-- ENTITIES    -->\n",fd_out);
		fputs("<!ENTITY % doctype \"",fd_out);
		fputs(this_elem->name,fd_out);
		fputs("\"	-- Document type GI -->\n\n",fd_out);
			/*
			| Check to see if there any Entitites before continuing
			*/
		if(!entity0){		/* No entity entries */
			return;
		}
	}
	FOREVER{
		build_entity_rec(this_ent);
		if(!this_ent->next_entity){  break;  }
		this_ent = this_ent->next_entity;
	}
}
/***********************+----------------+
|			| write_elements |
|			+----------------+
*/
void write_elements()
{
	struct element *this_elem;
	char           *content;
	char           *temp;

	this_elem = element0;

	fputs("\n                <!-- ELEMENTS    -->\n",fd_out);
	fputs("<!--      ELEMENTS       MIN  CONTENT -->\n",fd_out);
		/*
		| Change the name of the first element because it
		| is really the doc type. The reallocation of name
		| is necessary in case the original element0 name
		| was shorter than %doctype;.
		*/
	temp            = (char*) malloc(strlen("\%doctype;")+2);
	strcpy(temp,"\%doctype;");
	if(this_elem != NULL){
		free(this_elem->name);		/* give back old memory            */
		this_elem->name = temp;		/* replace orig element with entity*/
		/*
		| Loop through the element chain, writing the <!ELEMENT
		| statement for each entry. Entries after the first
		| occurrance of a tag are used for content models and do
		| not have <!ELEMENT statements created. This prevents
		| duplicate tags.
		*/
		FOREVER{
			if(this_elem->type != MULTIPLE){
				build_element_rec(this_elem);
			}
			if(!this_elem->next_element){
				break;
			}
			this_elem = this_elem->next_element;
		}
	}
}
/***********************+------------------+
|			| build_attrib_rec |
|			+------------------+
*/
void build_attrib_rec(struct attribute *this_att)
{
	int  i;
	int  col = tag_width + 4;
	int  col2;
	int  len;
	int  len2;
	int  status;
	static int  not_first;
	static char curr_element[256];

		/*
		| Check to see if this is the same element
		*/
	status = strcmp(this_att->element,curr_element);
	if(status){
		strcpy(curr_element,this_att->element);
			/*
			| New element, write the element portion
			*/
		if(not_first){
			strcpy(out_buff,">\n");
			fputs(out_buff,fd_out);
		}
		not_first = TRUE;
		strcpy(out_buff,"<!ATTLIST ");
		strcat(out_buff,this_att->element);
		strcat(out_buff,"\n");
		fputs(out_buff,fd_out);
		out_buff[0] = '\0';
	}
		/*
		| Tab over so everything lines up
		*/
	strcpy(out_buff,"                         ");
	len  = strlen(this_att->name);
	len  = col - len + 1;
	strcat(out_buff,this_att->name);

	for(i=0;i<len;i++){
		strcat(out_buff," ");
	}
	strcat(out_buff,this_att->values);

	len2 = strlen(this_att->values);
	len2 = 15 - len2;
	for(i=0;i<len2;i++){
		strcat(out_buff," ");
	}
	strcat(out_buff,this_att->default_value);
	strcat(out_buff,"\n");

	fputs(out_buff,fd_out);
	out_buff[0] = '\0';
}
/***********************+------------------+
|			| write_attributes |
|			+------------------+
*/
void write_attributes()
{
	struct attribute *this_att;

	fputs("\n",fd_out);
	fputs("                <!-- ATTRIBUTES  -->\n",fd_out);
	fputs("<!--      ELEMENT        NAME         VALUE          DEFAULT -->\n",fd_out);

	if(attribute0 == 0){
		return; }
	printf("\n");
	this_att = attribute0;
	FOREVER{
		build_attrib_rec(this_att);
		if(!this_att->next_attribute){  break;  }
		this_att = this_att->next_attribute;
	}
	strcpy(out_buff,">\n");
	fputs(out_buff,fd_out);
}
/***********************+--------------------+
|			| write_default_decl |
|			+--------------------+
*/
void write_default_decl()
{
	fputs("<!SGML  \"ISO 8879:1986\"\n",fd_out);
	fputs("-- This default SGML declaration inserted by EasyDTD. --\n",fd_out);
	fputs("CHARSET\n",fd_out);
	fputs("         BASESET  \"ISO 646-1983//CHARSET\n",fd_out);
	fputs("                   International Reference Version (IRV)//ESC 2/5 4/0\"\n",fd_out);
	fputs("         DESCSET  0   9   UNUSED\n",fd_out);
	fputs("                  9   2   9\n",fd_out);
	fputs("                  11  2   UNUSED\n",fd_out);
	fputs("                  13  1   13\n",fd_out);
	fputs("                  14  18  UNUSED\n",fd_out);
	fputs("                  32  95  32\n",fd_out);
	fputs("                  127 1   UNUSED\n",fd_out);
	fputs("CAPACITY        SGMLREF\n",fd_out);
	fputs("                TOTALCAP        150000\n",fd_out);
	fputs("                GRPCAP          150000\n",fd_out);
	fputs("SCOPE    DOCUMENT\n",fd_out);
	fputs("SYNTAX   \n",fd_out);
	fputs("         SHUNCHAR CONTROLS 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18\n",fd_out);
	fputs("                           19 20 21 22 23 24 25 26 27 28 29 30 31 127 255\n",fd_out);
	fputs("         BASESET  \"ISO 646-1983//CHARSET\n",fd_out);
	fputs("                   International Reference Version (IRV)//ESC 2/5 4/0\"\n",fd_out);
	fputs("         DESCSET  0 128 0\n",fd_out);
	fputs("         FUNCTION RE          13\n",fd_out);
	fputs("                  RS          10\n",fd_out);
	fputs("                  SPACE       32\n",fd_out);
	fputs("                  TAB SEPCHAR  9\n",fd_out);
	fputs("         NAMING   LCNMSTRT \"\"\n",fd_out);
	fputs("                  UCNMSTRT \"\"\n",fd_out);
	fputs("                  LCNMCHAR \".-\"\n",fd_out);
	fputs("                  UCNMCHAR \".-\"\n",fd_out);
	fputs("                  NAMECASE GENERAL YES\n",fd_out);
	fputs("                           ENTITY  NO\n",fd_out);
	fputs("         DELIM    GENERAL  SGMLREF\n",fd_out);
	fputs("                  SHORTREF SGMLREF\n",fd_out);
	fputs("         NAMES    SGMLREF\n",fd_out);
	fputs("         QUANTITY SGMLREF\n",fd_out);
	fputs("                  NAMELEN  32\n",fd_out);
	fputs("                  TAGLVL   24\n",fd_out);
	fputs("                  LITLEN   1024\n",fd_out);
	fputs("                  GRPGTCNT 150\n",fd_out);
	fputs("                  GRPCNT   64\n",fd_out);
	fputs("FEATURES\n",fd_out);
	fputs("  MINIMIZE    DATATAG  NO    OMITTAG  YES   RANK     NO    SHORTTAG YES\n",fd_out);
	fputs("  LINK        SIMPLE   NO    IMPLICIT NO    EXPLICIT NO  \n",fd_out);
	fputs("  OTHER       CONCUR   NO    SUBDOC   NO    FORMAL   YES\n",fd_out);
	fputs("  APPINFO     NONE\n",fd_out);
	fputs(">\n",fd_out);
}
/***********************+--------------------+
|			| write_skeleton_dtd |
|			+--------------------+
*/
void write_skeleton_dtd()
{
/*
	struct element *this_elem;

	this_elem = element0;
*/
	if(SGML_declaration){
		write_default_decl();
	}
	write_dtd_header();
	write_entities();
	write_elements();
	write_attributes();
}
/***********************+--------------+
|			| set_rec_type |
|			+--------------+
*/
long set_rec_type()
{
	int status;

	status = strncmp(in_buff,"<ELEMENTS>",10);
	if(!status){			/* There is a match... */
		return((long)ELEMENT);
	}
	status = strncmp(in_buff,"<ENTITIES>",10);
	if(!status){			/* There is a match... */
		return((long)ENTITY);
	}
	status = strncmp(in_buff,"<ATTRIBUTES>",12);
	if(!status){			/* There is a match... */
		return((long)ATTRIBUTE);
	}
		/*
		| At this point, the command is invalid!!
		*/
/*
 * fprintf(stderr,"INVALID Record Type!\n");
*/
   return((long)ERROR);
}
/********************************************************************
=====================================================================
=                      DATA DICTIONARY CODE                         =
=                                                                   =
= This section of code writes a skeleton data dictionary for the    =
= DTD. The entities are written first, followed by the elements.    =
= The data dictionary DTD is defined in DATADICT.DTD, which is      =
= included in the EasyDTD distribution. The content model is also   =
= filled in.                                                        =
=                                                                   =
=====================================================================
*********************************************************************/
/***********************+-------------------+
|			| write_dd_entities |
|			+-------------------+
*/
#ifdef IGNORE
void write_dd_entities()
{
	struct entity *this_ent;

	if(entity0 == 0){  return; }
	printf("\nENTITIES:\n");
	this_ent = entity0;
	FOREVER{
		printf("%s %s\n",this_ent->name,this_ent->value);
		if(!this_ent->next_entity){  break;  }
		this_ent = this_ent->next_entity;
	}
}
#endif
/***********************+------------------+
|			| write_dd_attribs |
|			+------------------+
*/
void write_dd_attribs(struct element *this_element)
{
	int  first_time = TRUE;
	struct attribute *this_att;
	struct attribute *attrib;

	this_att = attribute0;		/* Start at the beginning... */
	FOREVER{
		attrib = lookup_elem_attribute(this_element->name,this_att);
		if(first_time){
			first_time = FALSE;
			if(!attrib){	/* No attributes for this element */
				if(minimization){
					fprintf(fd_out,"<ATTRIB>N/A\n");
				}else{
					fprintf(fd_out,"<ATTRIB>N/A</ATTRIB>\n");
				}
				return;
			}else{
				if(minimization){
					fprintf(fd_out,"<ATTRIB>%s\n",
						attrib->name);
				}else{
					fprintf(fd_out,"<ATTRIB>%s<ATTRIB>\n",
						attrib->name);
				}
			}
		}else{
			if(!attrib){	/* No more attributes */
					/*
					| Well, There's really one left!!
					*/
				if(minimization){
					fprintf(fd_out,"<ATTRIB>%s\n",
						this_att->name);
				}else{
					fprintf(fd_out,"<ATTRIB>%s<ATTRIB>\n",
						this_att->name);
				}
				break;
			}else{
				if(minimization){
					fprintf(fd_out,"<ATTRIB>%s\n",
						attrib->name);
				}else{
					fprintf(fd_out,"<ATTRIB>%s<ATTRIB>\n",
						attrib->name);
				}
			}
		}
		this_att = attrib->next_attribute;
	}
}
/***********************+-------------------+
|			| write_dd_elements |
|			+-------------------+
*/
void write_dd_elements()
{
	struct element *this_elem;
	struct attribute *this_attrib;
	int    len;
	int    truth;
	char content[MAX_CONTENT];

	if(element0 == 0){   return;  }
	this_elem = element0;
	FOREVER{
		if(!this_elem->next_element){  break;  }
			/*
			| if this element is one of the 'special' SGML
			| names, don't put out a record for it 
			*/
		truth = validate_lookup_table(content_model_table,this_elem->name);
		if(truth){
			this_elem = this_elem->next_element;
			continue;
		}
		if(verbose){
			printf("%s \n",this_elem->name);
		}
		fputs("<ELEMENT>\n",fd_out);
		strcpy(out_buff,"<TAG>");
		strcat(out_buff,this_elem->name);
		if(minimize){
			strcat(out_buff,"\n");
		}else{
			strcat(out_buff,"</TAG>\n");
		}
		fputs(out_buff,fd_out);
		fputs("<NAME>\n",fd_out);
		if(!minimize){
			fputs("</NAME>\n",fd_out);
		}
		fputs("<DESC>\n",fd_out);
		if(!minimize){
			fputs("</DESC>\n",fd_out);
		}
		fputs("<CONTENT>\n",fd_out);
		if(!minimize){
			fputs("</CONTENT>\n",fd_out);
		}
		strcpy(out_buff,"<MODEL>");
			/*
			| content model goes here...
			*/

		create_content_model(this_elem,content);
		len     = strlen(content);
		*(content + len - 1) = '\0';
		strcat(content,")");
		strcat(out_buff,content);
		fputs(out_buff,fd_out);
		if(minimize){
			fputs("\n",fd_out);
		}else{
			fputs("</MODEL>\n",fd_out);
		}
			/*
			| Attributes go here
			*/
		write_dd_attribs(this_elem);
			/*
			| Looks like a simple example could
			| automatically be generated for here for
			| a future release.
			*/
		fputs("<EXAMPLE>\n",fd_out);
		if(!minimize){
			fputs("</EXAMPLE>\n",fd_out);
		}
		fputs("<SEEALSO>\n",fd_out);
		if(!minimize){
			fputs("</SEEALSO>\n",fd_out);
		}
		fputs("</ELEMENT>\n",fd_out);
		if(!this_elem->next_element){  break;  }
		this_elem = this_elem->next_element;
	}
}
/***********************+----------+
|			| write_dd |
|			+----------+
*/
void write_dd()
{
	fputs("<DATA-DICTIONARY>\n",fd_out);
/*
Add this in a future release
	write_dd_entities();
*/
	write_dd_elements();
	fputs("</DATA-DICTIONARY>\n",fd_out);
}
/********************************************************************
=====================================================================
=                            .APP CODE                              =
=                                                                   =
= This section of code writes a skeleton .APP file. The file        =
= both variable and macro definitions to process a document for this=
= DTD with S-Engine. It is a convenience program that must be       =
= customized for whatever application you are writing.              =
=                                                                   =
=====================================================================
*********************************************************************/
/***********************+------------------+
|			| write_app_header |
|			+------------------+
| This function writes a set of comments at the top of the ouput
| file.
*/
void write_app_header()
{
	fputs("\n",fd_out);
	fputs("( FILE:                                          )\n",fd_out);
	fputs("( This skeleton .APP file was created by EASYDTD.)\n",fd_out);
	fputs("( It is a starting point with all of the         )\n",fd_out);
	fputs("( elements and attributes defined via EASYDTD.   )\n",fd_out);
	fputs("( If you also build your DTD via EASYDTD, this   )\n",fd_out);
	fputs("( .APP file is the starting point for S-Engine   )\n",fd_out);
	fputs("( applications for that DTD.                     )\n",fd_out);
	fputs("(                                                )\n",fd_out);
}
/***********************+----------------+
|			| write_attrib_v |
|			+----------------+
*/
void write_attrib_v(struct attribute *this_att)
{
		/*
		| Check to see if this is the same element
		*/
	strcpy(out_buff,"256 string ");
	strcat(out_buff,this_att->element);
	strcat(out_buff,".");
	strcat(out_buff,this_att->name);
	strcat(out_buff,"\n");
	fputs(out_buff,fd_out);
	out_buff[0] = '\0';
}
/***********************+-------------------+
|			| write_attrib_vars |
|			+-------------------+
*/
void write_attrib_vars()
{
	struct attribute *this_att;

	fputs("(       Write Attribute variables    )\n",fd_out);

	if(attribute0 == 0){
		return; }
	printf("\n");
	this_att = attribute0;
	FOREVER{
		write_attrib_v(this_att);
		if(!this_att->next_attribute){  break;  }
		this_att = this_att->next_attribute;
	}
}
/***********************+-------------------+
|			| write_element_var |
|			+-------------------+
*/
void write_element_var(struct element *this_elem)
{
		/*
		| Write the macro name
		*/
	strcpy(out_buff,"256 string ");
	strcat(out_buff,this_elem->name);
	strcat(out_buff," \n");
	fputs(out_buff,fd_out);
}
/***********************+----------------+
|			| write_app_vars |
|			+----------------+
| This file writes the base elements (ones defined by #PCDATA) as
| variables and the attributes as variables. The default size is
| 256 bytes.
*/
void write_app_vars()
{
	struct element *this_elem;
	char           *content;

	this_elem = element0;

	fputs("\n(        Variables for the elements. )\n",fd_out);
		/*
		| Loop through table, writing entries for each element
		*/
	FOREVER{
		if(this_elem->type != MULTIPLE){
			write_element_var(this_elem);
		}
		if(!this_elem->next_element){  break;  }
		this_elem = this_elem->next_element;
	}
}
/***********************+-----------------+
|			| write_start_def |
|			+-----------------+
*/
void write_start_def(struct element *this_elem)
{
		/*
		| Write the macro name
		*/
	strcpy(out_buff,": <");
	strcat(out_buff,this_elem->name);
	strcat(out_buff,"> \t");
	fputs(out_buff,fd_out);
	out_buff[0] = '\0';
		/*
		| Write the command to read the tag
		*/
	strcpy(out_buff,"\t");
	strcat(out_buff,this_elem->name);
	strcat(out_buff," read_into  ;\n");
	fputs(out_buff,fd_out);
}
/***********************+---------------+
|			| write_end_def |
|			+---------------+
*/
void write_end_def(struct element *this_elem)
{
		/*
		| Write the macro name
		*/
	strcpy(out_buff,": </");
	strcat(out_buff,this_elem->name);
	strcat(out_buff,"> \n");
	fputs(out_buff,fd_out);
	out_buff[0] = '\0';
		/*
		| Write the command to read the tag
		*/
	strcpy(out_buff,"\t");
	strcat(out_buff,this_elem->name);
	strcat(out_buff," display cr  ;\n");
	fputs(out_buff,fd_out);

}
/***********************+----------------+
|			| write_app_defs |
|			+----------------+
| Writes definitions to the output file for each element. The start
| element does a READ_INTO and the end element does a DISPLAY.
*/
void write_app_defs()
{
	struct element *this_elem;
	char           *content;

	this_elem = element0;

	fputs("\n(        Starting processing for the elements. )\n",fd_out);
		/*
		| Loop through table, writing entries for each element
		*/
	FOREVER{
		if(this_elem->type != MULTIPLE){
			write_start_def(this_elem);
			write_end_def(this_elem);
		}
		if(!this_elem->next_element){  break;  }
		this_elem = this_elem->next_element;
	}
}
/***********************+----------------+
|			| write_app_file |
|			+----------------+
*/
void write_app_file()
{
	write_app_header();
	write_app_vars();
	write_attrib_vars();
	write_app_defs();
}
/********************************************************************
=====================================================================
=                            AWK CODE                               =
=                                                                   =
= This section of code writes a skeleton .AWK file. The file        =
= contains the gsub calls to resolve entity references in the output=
= file.                                                             =
=                                                                   =
=====================================================================
*********************************************************************/
/***********************+------------------+
|			| awk_write_header |
|			+------------------+
| This function writes a set of comments at the top of the ouput
| file.
*/
void awk_write_header()
{
	fputs("\n",fd_out);
	fputs("# FILE: \n",fd_out);
	fputs("# This skeleton .AWK file was created by EasyDTD.\n",fd_out);
	fputs("# It is a starting point with all of the \n",fd_out);
	fputs("# entity references to be resolved. \n",fd_out);
	fputs("# \n",fd_out);
	fputs("BEGIN{\n",fd_out);
	fputs("#\tprint \"This AWK program generated by EasyDTD\"; \n",fd_out);
	fputs("}\n",fd_out);
	fputs("{\n",fd_out);
}
/***********************+-------------------+
|			| awk_write_trailer |
|			+-------------------+
| This function writes a set of comments at the top of the ouput
| file.
*/
void awk_write_trailer()
{
	fputs("\tprint;\n",fd_out);
	fputs("}\n",fd_out);
}
/***********************+---------------+
|			| write_awk_rec |
|			+---------------+
*/
void write_awk_rec(struct entity *this_entity)
{
	if(type_output == AWK){
		fprintf(fd_out,"\tgsub(/&%s;/,%s,$0);\n",
					this_entity->name,
					this_entity->value);
	}else{
		fprintf(fd_out,"\tgsub(/%s/,%s,$0);\n",
					this_entity->name,
					this_entity->value);
	}
}
/***********************+-------------------+
|			| write_awk_program |
|			+-------------------+
| Only General Entities are written to the file because Parameter
| Entities apply only to the SGML document instance.
*/
void write_awk_program()
{
	struct entity *this_ent;

	if(entity0 == 0){  return; }
	awk_write_header();
	this_ent = entity0;
	FOREVER{
		if(this_ent->type == GENERAL){
			write_awk_rec(this_ent);
		}
		if(!this_ent->next_entity){  break;  }
		this_ent = this_ent->next_entity;
	}
	awk_write_trailer();
}
/********************************************************************
=====================================================================
=                         OMNIMARK CODE                             =
=                                                                   =
= This section of code writes a skeleton OmniMark program to parse  =
= the elements defined by the input file.                           =
=                                                                   =
=====================================================================
*********************************************************************/
/***********************+-------------------+
|			| init_omni_progarm |
|			+-------------------+
*/
void init_omni_program()
{
	fprintf(fd_out,"; \n");
	fprintf(fd_out,"; This OmniMark program generated by EasyDTD.\n");
	fprintf(fd_out,"; It may need modification before being run!\n");
}
/***********************+----------------------+
|			| om_write_output_line |
|			+----------------------+
*/
void om_write_output_line(struct attribute *this_att)
{
	strcpy(out_buff,"\t\tOUTPUT \"  ");
	strcat(out_buff,this_att->name);
	strcat(out_buff,"=\%\"\%v(");
	strcat(out_buff,this_att->name);
	strcat(out_buff,")\%\" \%n\"\n");
	fprintf(fd_out,out_buff);
}
/***********************+----------------+
|			| write_omni_rec |
|			+----------------+
*/
void write_omni_rec(struct element *this_element)
{
	int  first_time = TRUE;
	struct attribute *this_att;
	struct attribute *attrib;

	this_att = attribute0;		/* Start at the beginning... */
	FOREVER{
		attrib = lookup_elem_attribute(this_element->name,this_att);
		if(first_time){
			first_time = FALSE;
			if(!attrib){	/* No attributes for this element */
				fprintf(fd_out,";\tPROCESS ELEMENT %s\n",this_element->name);
				fprintf(fd_out,"; \n");
				fprintf(fd_out,"ELEMENT %s\n",this_element->name);
				fprintf(fd_out,"\tOUTPUT \"<\%q>\%sc</\%q>\%n\"\n");
				fprintf(fd_out,"; \n");
				return;
			}else{
				fprintf(fd_out,";\tPROCESS ELEMENT %s\n",
					this_element->name);
				fprintf(fd_out,"; \n");
				fprintf(fd_out,"ELEMENT %s\n",this_element->name);
				fprintf(fd_out,"\tOUTPUT \"<\%q\"\n");
				fprintf(fd_out,"\tDO WHEN ATTRIBUTE %s IS SPECIFIED\n",
					this_att->name);
				om_write_output_line(this_att);
				fprintf(fd_out,"\tDONE\n");
			}
		}else{
			if(!attrib){	/* No more attributes */
				fprintf(fd_out,"\tOUTPUT \">\%%sc\%%n\"\n");
				fprintf(fd_out,"; \n");
				break;
			}else{
				fprintf(fd_out,"\tDO WHEN ATTRIBUTE %s IS SPECIFIED\n",
					this_att->name);
					om_write_output_line(this_att);
				fprintf(fd_out,"\tDONE\n");
			}
		}
		this_att = attrib->next_attribute;
	}
}
/***********************+--------------------+
|			| write_omni_program |
|			+--------------------+
*/
void write_omni_program()
{
	struct element *this_elem;

	if(element0 == 0){   return;  }
	init_omni_program();
		/*
		| Loop through table of elements, writing a rule for each
		*/
	this_elem = element0;
	FOREVER{
		if(this_elem->type != MULTIPLE){
			write_omni_rec(this_elem);
		}
		if(!this_elem->next_element){  break;  }
		this_elem = this_elem->next_element;
	}
	fputs("; \n",fd_out);
}
/********************************************************************
=====================================================================
=                      MAIN PROCESSING CODE                         =
=                                                                   =
= This section of code is the startup and main program.             =
=                                                                   =
=====================================================================
*********************************************************************/
/***********************+----------------+
|			| extract_string |
|			+----------------+
| This function extracts the next token from the source string
| into the destination string. The token delimiter marks the last
| character. For example to extract up to a ", pass " as delim.
*/
int extract_string(char delim, char* source, char* string)
{
	int  i;
	char *start;
	char ch;

	start = string;
	FOREVER{
		ch        = *source++;
		if(ch    == delim){  break;  }
		*string++ = ch;
	}
	*string = '\0';
	return(strlen(start));
}
/***********************+-------------------+
|			| process.file.args |
|			+-------------------+
*/
void process_file_args(char *filename)
{
	int  i;
	int  len;
	static int not_first_time;

	static char drive[64];
	static char dir[256];
	static char file[128];
	static char ext[64];
/*
	char drive[64];
	char dir[256];
	char file[128];
	char ext[64];
*/
	if(not_first_time){
		goto Process_Output;
	}
	not_first_time = TRUE;
		/*
		| Split input filename
		*/
	strcpy(file_in,filename);
	fnsplit(file_in,drive,dir,file,ext);
	len = strlen(ext);
	if(!len){		/* there is no file extension */
		strcpy(ext,".ezy");
	}
		/*
		| Build up input filename from pieces
		*/
	strcpy(file_in,drive);
	strcat(file_in,dir);
	strcat(file_in,file);
	strcat(file_in,ext);
		/*
		| Split output filename, if it exists
		*/
	return;
Process_Output:
	if(!filename){		/* No file, build from input filename */
		strcpy(file_out,drive);
		strcat(file_out,dir);
		strcat(file_out,file);
		switch(type_output){
		case AWK:
		case AWK2:
			strcat(file_out,".awk");
			break;
		case DTD:
		case ENTITY_ONLY:
			strcat(file_out,".dtd");
			break;
		case APP_FILE:
			strcat(file_out,".app");
			break;
		case OMNIMARK:
			strcat(file_out,".xom");
			break;
		case DATA_DICT:
			strcat(file_out,".sgm");
			break;
		default:
			;
		}
	}else{
		strcpy(file_out,filename);
	}
}
/***********************+---------------------+
|			| parse_1char_options |
|			+---------------------+
*/
void process_1char_options(char *arg)
{
	int  len;
	char flag;
	char ch;
	char number[8];

	while(*arg){
		flag = *arg++;
		switch(flag){
		case 'm':		/* minimization flag */
		case 'M':
			strcpy(minimization," - o ");
			minimize        = TRUE;
			break;
		case 'p':		/* Para short ref flag */
		case 'P':
			short_ref       = TRUE;
			break;
		case 'n':		/* Write eNtity elements */
		case 'N':
			entity_elements = TRUE;
			break;
		case 'v':		/* verbose mode        */
		case 'V':
			verbose         = TRUE;
			break;
		case 'q':		/* Quiet mode          */
		case 'Q':
			quiet           = TRUE;
			break;
		case 'g':		/* Gen. Entities       */
		case 'G':
			entity_file     = TRUE;
			break;
		case 'e':		/* List elements       */
		case 'E':
			list_elem       = TRUE;
			break;
		case 'l':		/* Show long names     */
		case 'L':
			long_names      = TRUE;
			break;
		case 'o':		/* Default connector   */
		case 'O':		/*  Use next char only */
			ch   = *arg++;
			len  = extract_string(ch,arg,default_occurrance);
			arg += len + 1;
			break;
		case 's':		/* Default separator   */
		case 'S':
			ch   = *arg++;
			len  = extract_string(ch,arg,default_separator);
			arg += len + 1;
			break;
		case 't':		/* Max Tag length      */
		case 'T':
				/*
				| Parse off length
				| Note that only two digits are examined, max.
				*/
			ch        = *arg++;
			len       = extract_string(ch,arg,number);
			arg      += len + 1;
			tag_width = atoi(number);
			break;
		case 'a':		/* Gen .APP file for S-Engine */
		case 'A':
			type_output      = APP_FILE;
			break;
		case 'w':
		case 'W':
			type_output      = AWK;
			quiet            = TRUE;
			break;
		case 'z':
		case 'Z':
			type_output      = AWK2;
			quiet            = TRUE;
			break;
		case 'd':
		case 'D':
			type_output      = DTD;
			break;
		case 'k':
		case 'K':
			type_output      = OMNIMARK;
			break;
		case 'c':
		case 'C':
			SGML_declaration = TRUE;
			break;
		case 'y':
		case 'Y':
			type_output      = DATA_DICT;
			break;
		case 'x':
		case 'X':
			type_output      = ENTITY_ONLY;
			break;
		default:
			usage();
			exit(1);
		}
	}
}
/***********************+------------------------+
|			| parse_cmd_line_options |
|			+------------------------+
*/
void parse_cmd_line_options(int argc, char **argv)
{
	long success;
	int  i;
	int  file_calls = 0;
	char trash[128];
		/*
		| make sure there are enough args.
		*/
/*
	if(argc < 2){
		usage();
		exit(1);
	}
*/
		/*
		| Initialization and such
		*/
	for(i=1;i<argc;i++){
		strcpy(trash,argv[i]);
		if(trash[0] == '-'){		/* Got cmd line args */
				process_1char_options(&trash[1]);
/*
			switch(trash[1]){
			case 'f':
			case 'F':
				process_1char_options(argv[i]);
				break;
			case 'w':
			case 'W':
				type_output      = AWK;
				quiet            = TRUE;
				break;
			default:
				printf("Invalid flag: %c Ignored\n",trash[1]);
			}
*/
		}else{
			process_file_args(trash);
			file_calls++;
		}
	}
	switch(file_calls){
	case 0:
		printf("parse_cmd_line_options: No files specified!\n");
		usage();
		exit(1);
		break;
	case 1:				/* Input file, but no output file */
		process_file_args((char*)0);
		break;
	case 2:
		break;
	default:
		printf("parse_cmd_line_options: Too many files!\n");
		exit(1);
	}
}
/***********************+------+
|			| main |
|			+------+
*/
void main(int argc, char *argv[])
{
	long rec_type = ELEMENT;
	int  status;

	char *upper;
	char *not_done;
	char flag;
		/*
		| Check to see if there is a processing flag
		*/
	parse_cmd_line_options(argc,argv);

	init_easydtd();
	FOREVER{
	   not_done = read_next_line();
	   if(not_done){
		if(in_buff[0]   == '<'){
/*			upper    = strupr(in_buff);	/ * Convert to U/C for simpler proc. */
			rec_type = set_rec_type();
			continue;
		}
		switch(rec_type){
		case ELEMENT:
			load_element();
			break;
		case ENTITY:
			load_entity();
			break;
		case ATTRIBUTE:
			load_attribute();
			break;
		case SGML_DEC:
		case INCLUDE:
		case COMMENT:
		default:
			load_entity();
/*
			fprintf(stderr,"INVALID record type!\n");
*/
			break;
		}
	   }else{
		break;
	   }
	}
		/*
		| At this point, the linked list of element structures
		| has been built and is ready to be used. There are
		| several programs that I have in mind that will
		| use this as their starting point.
		*/
		/*
		| For debug purposes, dump the list...
		*/
	if(verbose){
		dump_elements();
		dump_entities();
		dump_attributes();
	}
		/*
		| Dump the elements to the output file.
		*/
	switch(type_output){
	case APP_FILE:
		write_app_file();	/* Gen .APP file 	*/
		break;
	case DTD:
		write_skeleton_dtd();
		exit_easydtd();
		break;
	case AWK:
	case AWK2:
		write_awk_program();
		break;
	case OMNIMARK:
		write_omni_program();
		break;
	case DATA_DICT:
		write_dd();
		break;
	case ENTITY_ONLY:
		write_entities();
		break;
	default:
		;
	}
	fclose(fd_out);
	if(entity_file){
		fclose(fd_entity);
	}
	printf("End of run.\n");
}
